from dataclasses import dataclass
from lcwc.incident import Incident
from collections import namedtuple


@dataclass
class Coordinates:
    """Represents the coordinates of an incident"""

    """ The longitude of the incident """
    longitude: float

    """ The latitude of the incident """
    latitude: float


@dataclass
class ArcGISIncident(Incident):
    """Represents an incident from the live ArcGIS REST API"""

    """ The number of the incident """
    number: int

    """ The priority of the incident """
    priority: int

    """ The agency handling the incident """
    agency: str

    """ Whether the incident is public """
    public: bool

    """ The coordinates of the incident """
    coordinates: Coordinates



import datetime
import aiohttp
import pytz
from bs4 import BeautifulSoup
from lcwc import Client
from lcwc.category import IncidentCategory
from lcwc.unit import Unit
from lcwc.web.incident import WebIncident as Incident

DATE_FORMAT = "%a, %b %d, %Y %H:%M"
""" The date format used on the LCWC website """


class WebClient(Client):
    """Client for scraping the live incident page"""

    URL = "https://www.lcwc911.us/live-incident-list"
    """ The URL of the live incident page """

    @property
    def name(self) -> str:
        """Returns the name of the client"""
        return "WebClient"

    async def get_incidents(
        self, session: aiohttp.ClientSession, timeout: int = 10
    ) -> list[Incident]:
        """Fetches the live incident page and returns a list of incidents

        :param session: The aiohttp session to use
        :param timeout: The timeout in seconds
        :return: A list of incidents
        :rtype: list[Incident]
        """
        html = None
        async with session.get(self.URL, timeout=timeout) as resp:
            if resp.status == 200:
                html = await resp.read()
            else:
                raise Exception(f"Unable to fetch live incident page: {resp.status}")

        active_incidents = self.__parse(html)
        return active_incidents

    def __parse(self, page_html: bytes) -> list[Incident]:
        """Parses the live incident page and returns a list of incidents

        :param page_html: The html of the live incident page
        :return: A list of incidents
        :rtype: list[Incident]
        """

        incidents = []

        soup = BeautifulSoup(page_html, "html.parser")
        containers = soup.find_all("div", class_="live-incident-container")

        for container in containers:
            header = container.find("h2").text
            category = IncidentCategory[header.split()[1].upper()]

            table = container.find("table", class_="live-incidents")

            rows = table.find_all("tr")

            for row in rows:
                date_row = row.find("td", class_="date-row")
                incident_row = row.find("td", class_="incident-row")
                location_row = row.find("td", class_="location-row")
                units_row = row.find("td", class_="units-row")

                if (
                    not date_row
                    or not incident_row
                    or not location_row
                    or not units_row
                ):
                    continue

                # convert date to UTC
                local_tz = pytz.timezone("America/New_York")
                raw_date = datetime.datetime.strptime(
                    date_row.text.strip(), DATE_FORMAT
                )
                local_dt = local_tz.localize(raw_date, is_dst=None)
                date = local_dt.astimezone(pytz.utc)

                description = incident_row.text.strip().strip()

                # split location by street(s) and municipality (if applicable)
                location = [l.strip() for l in location_row.text.strip().split("\n")]

                if len(location) == 1:
                    intersection = None
                    municipality = location[0]
                else:
                    intersection = location[0]
                    municipality = location[1]

                # we have to decode manually because of internal <br/> tags
                unit_names = [
                    u.strip()
                    for u in units_row.decode_contents().strip().split("<br/>")
                ]

                units = [Unit(u) for u in unit_names if u != ""]

                incident = Incident(
                    category, date, description, municipality, intersection, units
                )
                incidents.append(incident)

        return incidents



from lcwc.category import IncidentCategory
from lcwc.unit import Unit

""" A list of keywords that indicate a location name"""
FIRE_UNIT_NAMES = [
    "ATV",
    "BATTALION",
    "BOAT",
    "BRUSH",
    "CHIEF",
    "DEPUTY",
    "DUTY OFFICER",
    "ENGINE",
    "FIRE POLICE",
    "FOAM",
    "RESCUE",
    "SQUAD",
    "TAC",
    "TRUCK",
    "UTILITY",
    "UTV",
]

""" A list of keywords that indicate a location name """
MEDICAL_UNIT_NAMES = ["AMB", "EMS", "INT", "MEDIC", "QRS"]

""" A list of keywords that indicate a location name """
LOCATION_NAMES = [
    "ALY",
    "AVE",
    "CIR",
    "CT",
    "DR",
    "LN",
    "PL",
    "PIKE",
    "RAMP",
    "RD",
    "ROUTE",
    "ST",
]

""" A list of keywords that indicate a medical incident """
MEDICAL_DESCRIPTION_KEYWORDS = ["MEDICAL"]

""" A list of keywords that indicate a fire incident """
FIRE_DESCRIPTION_KEYWORDS = ["FIRE"]

""" A list of keywords that indicate a traffic incident """
TRAFFIC_DESCRIPTION_KEYWORDS = ["TRAFFIC", "VEHICLE"]


def determine_category(description: str, units: list[Unit]) -> IncidentCategory:
    """Determines the category of an incident based on the description and units assigned

    :param description: The description of the incident
    :param units: The units assigned to the incident
    :return: The category of the incident
    :rtype: IncidentCategory
    """

    # check for unit matches
    for unit in units:
        if any(k in unit.name for k in FIRE_UNIT_NAMES):
            return IncidentCategory.FIRE
        elif any(k in unit.name for k in MEDICAL_UNIT_NAMES):
            return IncidentCategory.MEDICAL

    # extra note regarding traffic incidents: they tend to not have units assigned
    # unless there is an accompanying fire or medical incident for the same call
    # this needs to be checked before the description check for other categories
    if len(units) == 0 and any(k in description for k in TRAFFIC_DESCRIPTION_KEYWORDS):
        return IncidentCategory.TRAFFIC

    # perform a basic description check
    if any(k in description for k in MEDICAL_DESCRIPTION_KEYWORDS):
        return IncidentCategory.MEDICAL
    if any(k in description for k in FIRE_DESCRIPTION_KEYWORDS):
        return IncidentCategory.FIRE

    return IncidentCategory.UNKNOWN



import aiohttp
import datetime
import feedparser
import pytz

from lcwc import Client
from lcwc.unit import Unit
from .utils import (
    FIRE_UNIT_NAMES,
    LOCATION_NAMES,
    MEDICAL_UNIT_NAMES,
    determine_category,
)
from lcwc.feed.incident import FeedIncident

"""
Example entry:

<item>
    <title>MEDICAL EMERGENCY</title>
    <link>http://www.lcwc911.us/lcwc/lcwc/publiccad.asp</link>
    <description>MARTIC TOWNSHIP; BRIDGE VALLEY RD & LAKE ALDRED TER; MEDIC 56-1; </description>
    <pubDate>Wed, 25 Jan 2023 15:22:54 GMT</pubDate>
    <guid isPermaLink="false">792d7e14-34ef-4907-bb50-86c43cd3d570</guid>
</item>
"""


class FeedClient(Client):
    """Client for the incident RSS feed"""

    URL = "https://webcad.lcwc911.us/Pages/Public/LiveIncidentsFeed.aspx"
    """ The URL of the live incident feed """

    @property
    def name(self) -> str:
        """Returns the name of the client"""
        return "FeedClient"

    async def get_incidents(
        self, session: aiohttp.ClientSession, timeout: int = 10
    ) -> list[FeedIncident]:
        """Gets the live incident feed and returns a list of incidents

        :param session: The aiohttp session to use
        :param timeout: The timeout for the request
        :return: A list of incidents
        :rtype: list[Incident]
        """
        response = None
        async with session.get(self.URL, timeout=timeout) as resp:
            if resp.status == 200:
                response = await resp.read()
            else:
                raise Exception(f"Unable to fetch live incident feed: {resp.status}")

        active_incidents = self.__parse(response)
        return active_incidents

    def __parse(self, contents: bytes) -> list[FeedIncident]:
        """Parses the live incident feed and returns a list of incidents

        :param contents: The xml of the live incident feed
        :return: A list of incidents
        :rtype: list[Incident]
        """
        incidents = []

        d = feedparser.parse(contents)

        def has_intersection(details_segment: str) -> bool:
            return any(k in details_segment for k in LOCATION_NAMES)

        def has_unit_names(details_segment: str) -> bool:
            return any(
                k in details_segment for k in FIRE_UNIT_NAMES + MEDICAL_UNIT_NAMES
            )

        for entry in d.entries:
            guid = entry.guid
            gmt_date = datetime.datetime.strptime(
                entry.published, "%a, %d %b %Y %H:%M:%S %Z"
            )
            date = gmt_date.replace(tzinfo=datetime.timezone.utc).astimezone(pytz.utc)
            description = entry.title

            # Possible formats:
            # [municipality];[intersection];[units assigned]
            # [municipality];[intersection]
            # [municipality];[units assigned]
            # [municipality]
            details_split = entry["description"].strip().split(";", maxsplit=3)

            # first item is always the municipality
            municipality = details_split[0].strip()

            intersection = None
            units = []

            # skip if only municipality is present
            if len(details_split) >= 2:
                if has_unit_names(details_split[1]):
                    units = self.__extract_units(details_split[1])
                else:
                    if has_intersection(details_split[1]):
                        intersection = details_split[1].strip()
                    if len(details_split) > 2 and has_unit_names(details_split[2]):
                        units = self.__extract_units(details_split[2])

            category = determine_category(description, units)

            incidents.append(
                FeedIncident(
                    category, date, description, municipality, intersection, units, guid
                )
            )

        return incidents

    def __extract_units(self, units_data: str) -> list[str]:
        """Extracts the units from the data string

        :param units_data: The data string containing the units
        :return: A list of units
        :rtype: list[str]
        """
        units_data = units_data.strip()
        if units_data == "":
            return []
        unit_names = [u.strip() for u in units_data.strip().split("<br />")]
        return [Unit(u) for u in unit_names]



import logging
import aiohttp
import datetime
import json
import re
from lcwc import Client
from lcwc.arcgis.incident import ArcGISIncident, Coordinates
from lcwc.category import IncidentCategory
from lcwc.unit import Unit
from lcwc.utils.restadapter import RestAdapter, RestException


class ArcGISClient(Client):
    """Client for the ArcGIS REST API"""

    def __init__(self) -> None:
        super().__init__()
        self.logger = logging.getLogger(__name__)

    @property
    def name(self) -> str:
        return "ArcGISClient"

    async def get_incidents(
        self, session: aiohttp.ClientSession, timeout: int = 10
    ) -> list[ArcGISIncident]:
        """Fetches the page and parses the contents and returns a list of incidents"""

        layer_mapping = {
            IncidentCategory.FIRE: 0,
            IncidentCategory.MEDICAL: 1,
            IncidentCategory.TRAFFIC: 2,
        }

        # some fields are only present for certain incident types (ex: Priority is only present for medical and traffic)
        fields = {
            IncidentCategory.FIRE: [
                "IncidentNumber",
                "IncidentMunicipality",
                "IncidentOrigination",
                "PrimaryAgency",
                "CurrentUnits",
                "PublicLocation",
                "PublicType",
                "IsPublic",
            ],
            IncidentCategory.MEDICAL: [
                "IncidentNumber",
                "IncidentMunicipality",
                "IncidentOrigination",
                "PrimaryAgency",
                "CurrentUnits",
                "PublicLocation",
                "PublicType",
                "IsPublic",
                "Priority",
            ],
            IncidentCategory.TRAFFIC: [
                "IncidentNumber",
                "IncidentMunicipality",
                "IncidentOrigination",
                "PrimaryAgency",
                "CurrentUnits",
                "PublicLocation",
                "PublicType",
                "IsPublic",
                "Priority",
            ],
        }

        incidents = []

        for cat in IncidentCategory:
            if cat == IncidentCategory.UNKNOWN:
                continue

            if cat not in layer_mapping:
                self.logger.error(f"No layer mapping for {cat}")
                continue

            layer_id = layer_mapping[cat]

            """ Actual spatial extent of Lancaster County based LanCo GIS data
            lanco_spatial = {
                'xmin': -8548898.732776089,
                'ymin': 4845979.963808246,
                'xmax': -8432714.449782776,
                'ymax': 4909881.3194545675,
                'spatialReference': {
                    'wkid': 102100
                }
            }
            """

            # seems we need to expand the spatial extent to get all incidents
            lanco_spatial = {
                "xmin": -8657540.868810708,
                "ymin": 4794222.228992932,
                "xmax": -8290643.133041878,
                "ymax": 5048910.407239126,
                "spatialReference": {"wkid": 102100},
            }

            params = {
                "f": "json",
                "where": "1=1",
                "returnGeometry": "true",
                "spatialRel": "esriSpatialRelIntersects",
                "geometry": json.dumps(lanco_spatial),
                "geometryType": "esriGeometryEnvelope",
                "inSR": 102100,
                "outFields": ",".join(fields[cat]),
                "outSR": 4326,  # return coordinates in WGS84
                "currentTimestamp": int(
                    datetime.datetime.now().timestamp() * 1000
                ),  # add a timestamp to prevent caching
            }

            adapter = RestAdapter(
                session,
                "utility.arcgis.com",
                "usrsvcs/servers/a1f6aa7faab44b1582029509c46dce86/rest/services/Maps/Public_LiveFeeds/MapServer/",
            )

            try:
                resp = await adapter.get(endpoint=f"{layer_id}/query", ep_params=params)

            except RestException as e:
                self.logger.error(f"{cat} Error: {e}")
                continue

            self.logger.debug(f"{resp.url}")

            if resp.status_code != 200:
                self.logger.error(f"Error: {resp.status_code} for {cat}")
                continue

            error = resp.data.get("error", None)
            if error:
                self.logger.error(f"Response error: {error}")
                continue

            if "features" not in resp.data:
                continue

            for feature in resp.data["features"]:
                incident = self.__parse_incident(cat, feature)
                incidents.append(incident)

        return incidents

    def __parse_incident(
        self, category: IncidentCategory, incident: dict
    ) -> ArcGISIncident:
        attributes = incident["attributes"]
        geometry = incident["geometry"]

        date = datetime.datetime.fromtimestamp(attributes["IncidentOrigination"] / 1000)
        municipality = attributes["IncidentMunicipality"]

        intersection = re.sub(
            " +", " ", attributes["PublicLocation"]
        )  # collapse multiple spaces

        # unit names are condensed, lacking spaces and delimiters (ex: MED8611)
        if "CurrentUnits" in attributes and attributes["CurrentUnits"] is not None:
            unit_names = attributes["CurrentUnits"].split(",")
            units = [Unit(unit_name) for unit_name in unit_names]
        else:
            units = []

        number = int(attributes["IncidentNumber"])

        if "Priority" in attributes:
            priority = int(attributes["Priority"])
        else:
            priority = None
        agency = attributes["PrimaryAgency"]
        public = bool(attributes["IsPublic"])
        description = attributes["PublicType"]

        coords = Coordinates(geometry["x"], geometry["y"])

        incident = ArcGISIncident(
            category,
            date,
            description,
            municipality,
            intersection,
            units,
            number,
            priority,
            agency,
            public,
            coords
        )
        return incident



import datetime
from lcwc.web.incident import WebIncident as Incident

from .encoding import IncidentEncoder, IncidentDecoder


def is_related_incident(a: Incident, b: Incident, delta: datetime.timedelta) -> bool:
    """Determines if two incidents are related based on the intersection and time delta

    :param a: The first incident
    :param b: The second incident
    :param delta: The time delta
    :return: True if the incidents are related, False otherwise
    :rtype: bool
    """

    if a.intersection is None or b.intersection is None:
        return False
    return a.intersection == b.intersection and abs(a.date - b.date) <= delta



from dataclasses import dataclass


@dataclass(eq=False)
class Unit:

    """Represents a unit responding to an incident"""

    """ The name of the unit """
    name: str

    def __eq__(self, other):
        if not isinstance(other, Unit):
            return NotImplemented
        return self.name == other.name
